const axios = require("axios");
const crypto = require("crypto");

class TestResultReporter {
  constructor(config) {
    this.apiUrl = process.env.API_URL || "https://app.testresult.co/api/v1";
    this.apiKey = process.env.API_KEY;
    this.testSuiteUuid = process.env.TEST_SUITE_UUID;
    this.runId = process.env.TEST_RUN_ID || crypto.randomUUID();
    this.debugReport =
      (process.env.DEBUG_REPORT || "0") === "1" ||
      (process.env.DEBUG_REPORT === undefined && config && config.debugReport);
    this.muteConsole =
      process.env.MUTE_CONSOLE === "1" || (config && config.muteConsole);

    this.fullReport = {
      suites: [],
      errors: [],
      reportSlowTests: {},
      shardInfo: null,
    };
    this.projects = new Map();
    this.startTime = null;
    this.endTime = null;
    this.lifecycleTimings = {
      globalSetup: null,
      globalTeardown: null,
      beforeAll: {},
      afterAll: {},
      beforeEach: {},
      afterEach: {},
    };
    this.stepTimings = new Map();
    this.shardInfo = null;
  }

  // Runs when the test suite starts
  onBegin(config, suite) {
    this.startTime = new Date();
    this.fullReport = {
      suites: [],
      errors: [],
      reportSlowTests: {},
      shardInfo: null,
    };
    for (const project of config.projects) {
      this.projects.set(project.name, project);
    }

    // Extract reportSlowTests values from config
    if (config.reportSlowTests) {
      this.fullReport.reportSlowTests = {
        threshold: config.reportSlowTests.threshold,
        max: config.reportSlowTests.max,
      };
    }

    // Capture shard information if it exists
    if (config.shard) {
      this.shardInfo = {
        current: config.shard.current,
        total: config.shard.total,
        startTime: this.startTime.toISOString(),
        runId: this.runId,
      };
      this.fullReport.shardInfo = this.shardInfo;
    }
  }

  onTestBegin(test) {
    const projectName = test.parent.project().name;
    const testId = this.generateUniqueTestId(test);
    this.lifecycleTimings.beforeEach[testId] = { startTime: new Date() };
  }

  generateUniqueTestId(test) {
    const projectName = test.parent.project().name;
    const testPath = `${test.location.file}`;
    const testTitle = test.title;
    const uniqueString = `${this.testSuiteUuid}|${projectName}|${testPath}|${testTitle}`;

    // Generate a SHA-256 hash of the unique string
    return crypto.createHash("sha256").update(uniqueString).digest("hex");
  }

  onTestEnd(test, result) {
    const projectName = test.parent.project().name;
    const status = result.status.toUpperCase();
    const coloredStatus =
      status === "FAILED"
        ? `\x1b[31m${status}\x1b[0m` // Red for failed
        : status === "PASSED"
        ? `\x1b[32m${status}\x1b[0m` // Green for passed
        : status; // No color for other statuses

    if (!this.muteConsole) {
      console.log(
        `Test ended: ${test.title} (Project: ${projectName}) - ${coloredStatus}`
      );
    }

    const uniqueTestId = this.generateUniqueTestId(test);

    const suite = this.findOrCreateSuite(test.parent);
    suite.specs.push({
      id: uniqueTestId,
      title: test.title,
      ok: result.ok,
      tags: test.tags,
      file: test.location.file,
      line: test.location.line,
      column: test.location.column,
      project: projectName,
      tests: [
        {
          id: uniqueTestId,
          timeout: test.timeout,
          annotations: test.annotations,
          expectedStatus: test.expectedStatus,
          projectName: projectName,
          results: [this.serializeTestResult(result)],
        },
      ],
    });

    if (this.lifecycleTimings.beforeEach[uniqueTestId]?.startTime) {
      this.lifecycleTimings.beforeEach[uniqueTestId].endTime = new Date();
      this.lifecycleTimings.beforeEach[uniqueTestId].duration =
        this.lifecycleTimings.beforeEach[uniqueTestId].endTime -
        this.lifecycleTimings.beforeEach[uniqueTestId].startTime;
    }

    suite.specs[suite.specs.length - 1].lifecycleTimings = {
      beforeEach: this.lifecycleTimings.beforeEach[uniqueTestId]?.duration,
      afterEach: this.lifecycleTimings.afterEach[uniqueTestId]?.duration,
      beforeAll: this.lifecycleTimings.beforeAll[test.parent.title]?.duration,
      afterAll: this.lifecycleTimings.afterAll[test.parent.title]?.duration,
    };
  }

  onStepBegin(test, result, step) {
    const stepId = `${this.generateUniqueTestId(test)}-${step.title}`;
    this.stepTimings.set(stepId, { startTime: new Date() });
  }

  onStepEnd(test, result, step) {
    const stepId = `${this.generateUniqueTestId(test)}-${step.title}`;
    const timing = this.stepTimings.get(stepId);
    if (timing) {
      timing.endTime = new Date();
      timing.duration = timing.endTime - timing.startTime;
      step.duration = timing.duration;
    }
  }

  onError(error) {
    this.fullReport.errors.push(error);
  }

  async onEnd(result) {
    this.endTime = new Date();
    this.fullReport.result = result;
    this.fullReport.startTime = this.startTime.toISOString();
    this.fullReport.endTime = this.endTime.toISOString();
    this.fullReport.duration = this.endTime - this.startTime;

    // Update shard timing information
    // * Future feature *
    if (this.shardInfo) {
      this.shardInfo.endTime = this.endTime.toISOString();
      this.shardInfo.duration = this.endTime - this.startTime;
      this.fullReport.shardInfo = this.shardInfo;
    }

    // Write debug output if debugReport is enabled
    if (this.debugReport) {
      const debugDir = "playwright-debug";
      const timestamp = new Date().toISOString().replace(/[:.]/g, "-");
      const shardSuffix = this.shardInfo
        ? `-shard-${this.shardInfo.current}`
        : "";
      const filename = `test-results-${timestamp}${shardSuffix}.json`;

      try {
        // Create debug directory if it doesn't exist
        if (!require("fs").existsSync(debugDir)) {
          require("fs").mkdirSync(debugDir, { recursive: true });
        }

        // Write the full report to a JSON file
        require("fs").writeFileSync(
          `${debugDir}/${filename}`,
          JSON.stringify(
            {
              url: `${this.apiUrl}/test_suites/${this.testSuiteUuid}/test_runs.json`,
              payload: {
                test_run: {
                  results: this.fullReport,
                  started_at: this.startTime.toISOString(),
                  ended_at: this.endTime.toISOString(),
                  duration: this.fullReport.duration,
                  lifecycle_timings: {
                    globalSetup: this.lifecycleTimings.globalSetup?.duration,
                    globalTeardown:
                      this.lifecycleTimings.globalTeardown?.duration,
                  },
                  shard_info: this.shardInfo,
                },
              },
            },
            null,
            2
          )
        );
        console.log(`Debug output written to ${debugDir}/${filename}`);
      } catch (error) {
        console.error("Error writing debug output:", error);
      }
    }
    if (!this.apiUrl || !this.testSuiteUuid) {
      console.error("Error: apiUrl or testSuiteUuid is undefined");
      return;
    }

    // Modify the URL to include shard information if present
    // * Future feature *
    let url = `${this.apiUrl}/test_suites/${this.testSuiteUuid}/test_runs.json`;
    if (this.shardInfo) {
      url += `?shard=${this.shardInfo.current}&total_shards=${this.shardInfo.total}&run_id=${this.runId}`;
    }

    console.log(`Attempting to send results to: ${url}`);

    const payload = {
      test_run: {
        results: this.fullReport,
        started_at: this.startTime.toISOString(),
        ended_at: this.endTime.toISOString(),
        duration: this.fullReport.duration,
        lifecycle_timings: {
          globalSetup: this.lifecycleTimings.globalSetup?.duration,
          globalTeardown: this.lifecycleTimings.globalTeardown?.duration,
        },
        shard_info: this.shardInfo,
      },
    };

    try {
      const response = await axios.post(url, payload, {
        headers: {
          "Content-Type": "application/json",
          "X-API-Key": this.apiKey,
        },
        timeout: 60000, // 60 second timeout
      });

      // Poll for completion if status is 'processing'
      if (response.data.status === "processing") {
        await this.pollForCompletion(response.data.id);
      }

      console.log("Test results processed successfully");
    } catch (error) {
      console.error("Error sending test results:", error.message);
    }
  }

  async pollForCompletion(testRunId) {
    const maxAttempts = 30;
    let attempts = 0;

    while (attempts < maxAttempts) {
      const response = await axios.get(
        `${this.apiUrl}/test_runs/${testRunId}/status`,
        {
          headers: { "X-API-Key": this.apiKey },
        }
      );

      if (["passed", "failed"].includes(response.data.status)) {
        return response.data;
      }

      attempts++;
      await new Promise((resolve) => setTimeout(resolve, 5000)); // Wait 5 seconds between polls
    }

    throw new Error(
      "Timeout waiting for test run processing. Check TestResult.co for results."
    );
  }

  findOrCreateSuite(suite) {
    if (!suite) return null;
    let found = this.fullReport.suites.find((s) => s.title === suite.title);
    if (!found) {
      found = { title: suite.title, specs: [] };
      this.fullReport.suites.push(found);
    }
    return found;
  }

  serializeTestResult(result) {
    return {
      duration: result.duration,
      status: result.status,
      error: result.error
        ? {
            message: result.error.message,
            stack: result.error.stack,
          }
        : undefined,
      stdout: result.stdout,
      stderr: result.stderr,
      retry: result.retry,
      steps: result.steps.map((step) => this.serializeTestStep(step)),
      attachments: result.attachments,
    };
  }

  serializeTestStep(step) {
    return {
      title: step.title,
      category: step.category,
      duration: step.duration,
      error: step.error
        ? {
            message: step.error.message,
            stack: step.error.stack,
          }
        : undefined,
      steps: step.steps.map((s) => this.serializeTestStep(s)),
    };
  }

  onGlobalSetup() {
    this.lifecycleTimings.globalSetup = { startTime: new Date() };
  }

  onGlobalTeardown() {
    this.lifecycleTimings.globalTeardown = { startTime: new Date() };
    // Calculate duration immediately since this is the last chance
    this.lifecycleTimings.globalTeardown.endTime = new Date();
    this.lifecycleTimings.globalTeardown.duration =
      this.lifecycleTimings.globalTeardown.endTime -
      this.lifecycleTimings.globalTeardown.startTime;
  }

  onBeforeAll(suite) {
    this.lifecycleTimings.beforeAll[suite.title] = { startTime: new Date() };
  }

  onAfterAll(suite) {
    if (this.lifecycleTimings.beforeAll[suite.title]?.startTime) {
      this.lifecycleTimings.afterAll[suite.title] = {
        startTime: new Date(),
        endTime: new Date(),
      };
      this.lifecycleTimings.afterAll[suite.title].duration =
        this.lifecycleTimings.afterAll[suite.title].endTime -
        this.lifecycleTimings.afterAll[suite.title].startTime;
    }
  }
}

module.exports = TestResultReporter;
